# -*- coding: utf-8 -*-
#
# This file is part of Calypso - CalDAV/CardDAV/WebDAV Server
# Copyright © 2016 Guido Günther <agx@sigxcpu.org>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Gssapi module.

This module handles kerberos authenticatien via gssapi
"""

import os
import calypso.acl
import calypso.config
import calypso.acl.nopwd

# pylint: disable=F0401
try:
    import kerberos as krb
except ImportError:
    krb = None
# pylint: disable=F0401


class Negotiate(object):
    _gssapi = False

    def __init__(self, log):
        self.log = log
        try:
            self.servicename = os.path.expanduser(config.get("server",
                                                             "servicename"))
        except:
            self.servicename = None

        if self.servicename and krb:
            self._gssapi = True

    def enabled(self):
        return self._gssapi

    def try_aaa(self, authorization, request, owner):
        """Perform authentication and authorization"""
        user, success = self.step(authorization, request)
        if success:
            return user, nopwd.has_right(owner, user, None)
        return user, False

    def step(self, authorization, request):
        """
        Try to authenticate the client and if succesful authenticate
        ourself to the client.
        """
        user = None

        if not self.enabled():
            return (None, False)

        try:
            (neg, challenge) = authorization.split()
            if neg.lower().strip() != 'negotiate':
                return (None, False)

            self.log.debug("Negotiate header found, trying Kerberos")
            result, context = krb.authGSSServerInit(self.servicename)
            result = krb.authGSSServerStep(context, challenge)

            if result == -1:
                return (None, False)

            response = krb.authGSSServerResponse(context)
            # Client authenticated successfully, so authenticate to the client:
            request.queue_header("www-authenticate",
                                 "negotiate " + response)
            user = krb.authGSSServerUserName(context)

            self.log.debug("Negotiate: found user %s" % user)
            result = krb.authGSSServerClean(context)
            if result != 1:
                self.log.error("Failed to cleanup gss context")
            return (user, True)
        except krb.GSSError as err:
            self.log.error("gssapi error: %s", err)

        return None, False



# -*- coding: utf-8 -*-
#
# This file is part of Calypso - CalDAV/CardDAV/WebDAV Server
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
# Copyright © 2011 Keith Packard
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Calypso collection classes.

Define the main classes of a collection as seen from the server.

"""

import os
import codecs
import time
import hashlib
import logging
import tempfile
import vobject
import re
import subprocess
import vobject.base
import iniparse
import configparser
import threading
from configparser import RawConfigParser as ConfigParser
from . import config, paths

METADATA_FILENAME = ".calypso-collection"

# pylint: enable=F0401

#
# Recursive search for 'name' within 'vobject'
#

def find_vobject_value(vobject, name):

    if vobject.name == name:
        return vobject.value

    for child in vobject.getChildren():
        value = find_vobject_value(child, name)
        if value:
            return value
    return None


class Item(object):

    """Internal item. Wraps a vObject"""

    def __init__(self, text, name=None, path=None, parent_urlpath=None):
        """Initialize object from ``text`` and different ``kwargs``."""

        self.log = logging.getLogger(__name__)

        # Preserve the original text as the contents of the object

        self.text = text

        # Strip out control characters

        text = re.sub(r"[\x01-\x09\x0b-\x1F\x7F]","", text)

        self.object = vobject.readOne(text)

        if 'x-calpyso-name' not in self.object.contents:
            if not name:
                if self.object.name == 'VCARD' or self.object.name == 'VEVENT':
                    if 'uid' not in self.object.contents:
                        self.object.add('UID').value = hashlib.sha1(text.encode('utf-8')).hexdigest()
                    name = self.object.uid.value
                else:
                    for child in self.object.getChildren():
                        if child.name == 'VEVENT' or child.name == 'VCARD':
                            if 'uid' not in child.contents:
                                child.add('UID').value = hashlib.sha1(text.encode('utf-8')).hexdigest()
                            name = child.uid.value
                            break
                    if not name:
                        name = hashlib.sha1(text.encode('utf-8')).hexdigest()

            self.object.add("X-CALYPSO-NAME").value = name
        else:
            names = self.object.contents[u'x-calypso-name']
            if len(names) > 1:
                self.object.contents[u'x-calypso-name'] = [names[0]]

        self.path = path
        self.name = self.object.x_calypso_name.value
        if parent_urlpath.endswith("/"):
            self.urlpath = parent_urlpath + self.name
        else:
            self.urlpath = "/".join([parent_urlpath, self.name])
        self.tag = self.object.name
        self.etag = hashlib.sha1(self.text.encode('utf-8')).hexdigest()

    @property
    def is_vcard(self):
        """Whether this item is a vcard entry"""
        if self.object.name == 'VCARD':
            return True
        if self.object.name == 'VEVENT':
            return False
        if self.object.name == 'VTODO':
            return False
        if self.object.name == 'VCALENDAR':
            return False
        for child in self.object.getChildren():
            if child.name == 'VCARD':
                return True
            if child.name == 'VEVENT':
                return False
            if child.name == 'VTODO':
                return False
        return False

    @property
    def is_vcal(self):
        """Whether this item is a vcal entry"""
        if self.object.name == 'VCARD':
            return False
        if self.object.name == 'VEVENT':
            return True
        if self.object.name == 'VTODO':
            return True
        if self.object.name == 'VCALENDAR':
            return True
        for child in self.object.getChildren():
            if child.name == 'VCARD':
                return False
            if child.name == 'VEVENT':
                return True
            if child.name == 'VTODO':
                return True
        return False

    @property
    def is_collection(self):
        return False

    @property
    def file_prefix(self):
        if self.is_vcard:
            return 'card-'
        if self.is_vcal:
            return 'cal-'
        return 'res-'

    @property
    def file_extension(self):
        if self.is_vcard:
            return '.vcf'
        if self.is_vcal:
            return '.ics'
        return '.dav'

    @property
    def length(self):
        return "%d" % len(self.text)

    @property
    def last_modified(self):
        value = find_vobject_value(self.object, "LAST-MODIFIED")
        if value:
            return value.utctimetuple()
        return time.gmtime()

    def match_name(self, name):
        if self.name == name:
            return True
        if self.name + self.file_extension == name:
            return True
        if self.name == name + self.file_extension:
            return True
        return False

    def __str__(self):
        fn = self.object.getChildValue("fn")
        if fn:
            return fn

        if hasattr(self.object, "vevent"):
            summary = self.object.vevent.getChildValue("summary")
            dtstart = self.object.vevent.getChildValue("dtstart")
            if summary and dtstart:
                return "%s (%s)"%(summary, dtstart)
            if summary:
                return summary
            if dtstart:
                return str(dtstart)

            uid = self.object.vevent.getChildValue("uid")
            if uid:
                return uid

        uid = self.object.getChildValue("uid")
        if uid:
            return uid

        return self.name

class Pathtime(object):
    """Path name and timestamps"""

    def __init__(self, path):
        self.path = path
        self.mtime = self.curmtime

    @property
    def curmtime(self):
        return os.path.getmtime(self.path)

    def is_up_to_date(self):
        newmtime = self.curmtime
        if newmtime == self.mtime:
            return True
        self.mtime = newmtime
        return False

class Directory(Pathtime):

    def __init__(self, path, name, parent_urlpath):
        Pathtime.__init__(self, path)
        self.name = name
        self.etag = hashlib.sha1(path.encode('utf-8')).hexdigest()
        if parent_urlpath.endswith("/"):
            self.urlpath = parent_urlpath + self.name
        else:
            self.urlpath = "/".join([parent_urlpath, self.name])

    @property
    def is_vcard(self):
        return False

    @property
    def is_vcal(self):
        return False

    @property
    def is_collection(self):
        return True

    @property
    def file_prefix(self):
        return ''

    @property
    def file_extension(self):
        return ''

    @property
    def ctag(self):
        return self.etag

    @property
    def color(self):
        return None

    def get_description(self):
        return self.urlpath

    @property
    def text(self):
        return self.path

    @property
    def length(self):
        return "%d" % len(self.path)

    @property
    def last_modified(self):
        return time.gmtime(self.curmtime)

    @property
    def is_calendar(self):
        return True

    @property
    def is_addressbook(self):
        return False

class CalypsoError(Exception):
    def __init__(self, name, reason):
        self.name = name
        self.reason = reason

    def __str__(self):
        return "%s: %s" % (self.reason, self.file)

class Collection(object):
    """Internal collection class."""

    def get_description(self):
        try:
            return str(self.metadata.get('collection', 'description'))
        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
            pass

        try:
            with codecs.open(os.path.join(self.path, ".git/description"), encoding='utf-8') as file:
                return file.read()
        except:
            # .git/description is not present eg when the complete server is a single git repo
            return self.urlpath

    def read_file(self, path):
        with codecs.open(path,encoding='utf-8') as file:
            text = file.read()
            item = Item(text, None, path, self.urlpath)
            return item

    def insert_file(self, path, name):
        if os.path.isdir(path):
            try:
                item = Directory(path, name, self.urlpath)
                self.my_dirs.append(item)
            except Exception as ex:
                self.log.exception("Insert %s failed", path)
                return
        else:
            try:
                item = self.read_file(path)
                self.my_items.append(item)
            except Exception as ex:
                self.log.exception("Insert %s failed", path)
                return

    def remove_file(self, path):
        old_items=[]
        for old_item in self.my_items:
            if old_item.path == path:
                old_items.append(old_item)
        for old_item in old_items:
            self.my_items.remove(old_item)
        old_dirs=[]
        for old_dir in self.my_dirs:
            if old_dir.path == path:
                old_dirs.append(old_dir)
        for old_dir in old_dirs:
            self.my_dirs.remove(old_dir)

    def scan_file(self, path, name):
        self.remove_file(path)
        self.insert_file(path, name)

    __metadatafile = property(lambda self: os.path.join(self.path, METADATA_FILENAME))

    def scan_metadata(self, force):
        try:
            mtime = os.path.getmtime(self.__metadatafile)
        except OSError:
            mtime = 0
            force = True

        if not force and mtime == self.mtime and self.metadata is not None:
            return

        parser = ConfigParser()
        parser.read(self.__metadatafile)
        self.metadata = parser

    def scan_dir(self, force):
        try:
            mtime = os.path.getmtime(self.path)
        except OSError:
            mtime = 0
            force = True

        self.scan_metadata(force)

        if not force and mtime == self.mtime:
            return
        self.log.debug("Scan %s", self.path)
        self.mtime = mtime
        filenames = os.listdir(self.path)
        newfiles = []
        for filename in filenames:
            if filename == METADATA_FILENAME:
                continue
            if filename == '.git':
                continue
            filepath = os.path.join(self.path, filename)
            for file in self.files:
                if filepath == file.path:
                    newfiles.append(file)
                    if not file.is_up_to_date():
                        self.log.debug("Changed %s", filepath)
                        self.scan_file(filepath, filename)
                    break
            else:
                self.log.debug("New %s", filepath)
                newfiles.append(Pathtime(filepath))
                self.insert_file(filepath, filename)

        for file in self.files:
            if os.path.basename(file.path) not in filenames:
                self.log.debug("Removed %s", file.path)
                self.remove_file(file.path)

        h = hashlib.sha1()
        self._text = ''
        for item in self.my_items:
            if getattr(item, 'etag', None):
                h.update(item.etag.encode('utf-8'))
            else:
                h.update(item.ctag.encode('utf-8'))
            self._text = self._text + item.text
        self._ctag = '%d-' % self.mtime + h.hexdigest()
        self.files = newfiles

    def __init__(self, path):
        """Initialize the collection with ``cal`` and ``user`` parameters."""

        self.log = logging.getLogger(__name__)
        self.encoding = "utf-8"
        self.urlpath = path
        self.owner = paths.url_to_owner(path)
        self.path = paths.url_to_file(path)
        self.files = []
        self.my_items = []
        self.my_dirs = []
        self.mtime = 0
        self._text = ''
        self._ctag = ''
        self.etag = hashlib.sha1(self.path.encode('utf-8')).hexdigest()
        self.metadata = None
        self.metadata_mtime = None
        self.scan_dir(False)
        self.tag = "Collection"
        self.lock = threading.Lock()

    def __str__(self):
        return "Calendar-%s (at %s)" % (self.name, self.path)

    def __repr__(self):
        return "<Calendar %s>" % (self.name)

    def acquire(self):
        self.lock.acquire()

    def release(self):
        self.lock.release()

    def __enter__(self):
        self.lock.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.__exit__(exc_type, exc_val, exc_tb)

    def has_git(self):
        return True

    def git_commit(self, context):
        args = ["git", "commit", "--allow-empty"]
        env = {}

        message = context.get('action', 'other action')

        if "user" in context:
            # use environment variables instead of --author to avoid git
            # looking it up in previous commits if it doesn't seem well-formed
            env['GIT_AUTHOR_NAME'] = context['user'] or "unknown"
            env['GIT_AUTHOR_EMAIL'] = "%s@webdav"%context['user']
            # supress a chatty message that we could configure author
            # information explicitly in the config file. (slicing it in after
            # the git command as position is important with git arguments)
            args[1:1] = ["-c", "advice.implicitIdentity=false"]
        if "user-agent" in context:
            message += u"\n\nUser-Agent: %r"%context['user-agent']

        args.extend(["-m", message.encode('utf8')])

        subprocess.check_call(args, cwd=self.path, env=env)

    def git_add(self, path, context):
        if self.has_git():
            subprocess.check_call(["git", "add", os.path.basename(path)], cwd=self.path)
            self.git_commit(context=context)

    def git_rm(self, path, context):
        if self.has_git():
            subprocess.check_call(["git", "rm", os.path.basename(path)], cwd=self.path)
            self.git_commit(context=context)

    def git_change(self, path, context):
        if self.has_git():
            subprocess.check_call(["git", "add", os.path.basename(path)], cwd=self.path)
            self.git_commit(context=context)
            # Touch directory so that another running instance will update
            try:
                os.utime(self.path, None)
            except Exception as ex:
                self.log.exception("Failed to set directory mtime")

    def write_file(self, item):
        fd, path = tempfile.mkstemp(item.file_extension, item.file_prefix, dir=self.path)
        self.log.debug('Trying to write to %s', path)
        file = os.fdopen(fd, 'w')
        file.write(item.text)
        file.close()
        self.log.debug('Wrote %s to %s', file, path)
        return path

    def create_file(self, item, context):
        # Create directory if necessary
        self.log.debug("Add %s", item.name)
        if not os.path.exists(os.path.dirname(self.path)):
            try:
                os.makedirs(os.path.dirname(self.path))
            except OSError as ose:
                self.log.exception("Failed to make collection directory %s: %s", self.path, ose)
                raise

        context['action'] = u'Add %s'%item

        try:
            path = self.write_file(item)
            self.git_add(path, context=context)
            self.scan_dir(True)
        except OSError as ex:
            self.log.exception("Error writing file")
            raise
        except Exception as ex:
            self.log.exception("Caught Exception")
            self.log.debug("Failed to create %s: %s", self.path,  ex)
            raise

    def destroy_file(self, item, context):
        self.log.debug("Remove %s", item.name)

        context['action'] = u'Remove %s'%item

        try:
            os.unlink(item.path)
            self.git_rm(item.path, context=context)
            self.scan_dir(True)
        except Exception as ex:
            self.log.exception("Failed to remove %s", item.path)
            raise

    def rewrite_file(self, item, context):
        self.log.debug("Change %s", item.name)

        context['action'] = u'Modify %s'%item

        try:
            new_path = self.write_file(item)
            os.rename(new_path, item.path)
            self.scan_file(item.path, item.name)
            self.git_change(item.path, context=context)
            self.scan_dir(True)
        except Exception as ex:
            self.log.exception("Failed to rewrite %s", item.path)
            raise

    def get_item(self, name):
        """Get collection item called ``name``."""
        for item in self.my_items:
            if item.match_name(name):
                return item
        return None

    def get_items(self, name):
        """Get collection items called ``name``."""
        items=[]
        for item in self.my_items:
            if item.match_name(name):
                items.append(item)
        return items

    def append(self, name, text, context):
        """Append items from ``text`` to collection.

        If ``name`` is given, give this name to new items in ``text``.

        """

        self.log.debug('append name %s', name)
        try:
            new_item = Item(text, name, None, self.urlpath)
        except Exception as e:
            self.log.exception("Cannot create new item")
            raise
        old_item = self.get_item(new_item.name)
        if old_item:
            self.log.debug("Item %s already present %s" , new_item.name, old_item.path)
            raise CalypsoError(new_item.name, "Item already present")
        self.log.debug("New item %s", new_item.name)
        self.create_file(new_item, context=context)
        return new_item

    def remove(self, name, context):
        """Remove object named ``name`` from collection."""
        self.log.debug("Remove object %s", name)
        for old_item in self.my_items:
            if old_item.match_name(name):
                self.destroy_file(old_item, context=context)

    def replace(self, name, text, context):
        """Replace content by ``text`` in objet named ``name`` in collection."""

        path=None
        old_item = self.get_item(name)
        if old_item:
            path = old_item.path

        try:
            new_item = Item(text, name, path, self.urlpath)
        except Exception:
            self.log.exception("Failed to replace %s", name)
            raise

        ret = False
        if path is not None:
            self.log.debug('rewrite path %s', path)
            self.rewrite_file(new_item, context=context)
        else:
            self.log.debug('remove and append item %s', name)
            self.remove(name)
            self.append(name, text, context=context)
        return new_item

    def import_item(self, new_item, path):
        old_item = self.get_item(new_item.name)
        if old_item:
            new_item.path = old_item.path
            self.rewrite_file(new_item, context={})
            self.log.debug("Updated %s from %s", new_item.name, path)
        else:
            self.create_file(new_item, context={})
            self.log.debug("Added %s from %s", new_item.name, path)

    def import_file(self, path):
        """Merge items from ``path`` to collection.
        """

        try:
            with codecs.open(path,encoding='utf-8') as file:
                new_object = vobject.readComponents(file.read())
                for new_ics in new_object:
                    if new_ics.name == 'VCALENDAR':

                        events = new_ics.vevent_list
                        for ve in events:
                            # Check for events with both dtstart and duration entries and
                            # delete the duration one
                            if 'dtstart' in ve.contents and 'duration' in ve.contents:
                                del ve.contents['duration']
                            new_ics.vevent_list = [ve]
                            new_item = Item(new_ics.serialize(), None, path, self.urlpath)
                            self.import_item(new_item, path)
                    else:
                        new_item = Item(new_ics.serialize(), None, path, self.urlpath)
                        self.import_item(new_item, path)
            return True
        except Exception as ex:
            self.log.exception("Failed to import: %s", path)
            return False

    def write(self, headers=None, items=None):
        return True

    @property
    def is_collection(self):
        return True

    @property
    def ctag(self):
        self.scan_dir(False)
        """Ctag from collection."""
        return self._ctag

    @property
    def name(self):
        """Collection name."""
        return self.path.split(os.path.sep)[-1]

    @property
    def text(self):
        """Collection as plain text."""
        self.scan_dir(False)
        return self._text

    @property
    def color(self):
        """Color."""
        try:
            return "#%s" % self.metadata.get('collection', 'color')
        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
            return None

    @property
    def headers(self):
        """Find headers items in collection."""
        return []

    @property
    def items(self):
        """Get list of all items in collection."""
        self.scan_dir(False)
        return self.my_items

    @property
    def dirs(self):
        """Get list of all dirs in collection."""
        self.scan_dir(False)
        return self.my_dirs

    @property
    def last_modified(self):
        """Get the last time the collection has been modified.

        The date is formatted according to rfc1123-5.2.14.

        """
        self.scan_dir(False)
        return time.gmtime(self.mtime)

    @property
    def length(self):
        return "%d" % len(self.text)

    @property
    def is_addressbook(self):
        try:
            return self.metadata.getboolean('collection', 'is-addressbook')
        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
            return True

    @property
    def is_calendar(self):
        try:
            return self.metadata.getboolean('collection', 'is-calendar')
        except (configparser.NoSectionError, configparser.NoOptionError, ValueError):
            return True



# -*- coding: utf-8 -*-
#
# This file is part of Calypso - CalDAV/CardDAV/WebDAV Server
# Copyright © 2011 Keith Packard
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Calypso Server module.

This module offers 3 useful classes:

- ``HTTPServer`` is a simple HTTP server;
- ``HTTPSServer`` is a HTTPS server, wrapping the HTTP server in a socket
  managing SSL connections;
- ``CollectionHTTPHandler`` is a WebDAV request handler for HTTP(S) servers.

To use this module, you should take a look at the file ``calypso.py`` that
should have been included in this package.

"""

import os
import os.path
import base64
import socket
import time
import email.utils
import logging
import email
import ssl
import threading
from http import client, server


from . import acl, config, webdav, xmlutils, paths, gssapi

log = logging.getLogger()
ch = logging.StreamHandler()
formatter = logging.Formatter("%(message)s")
ch.setFormatter (formatter)
log.addHandler(ch)
negotiate = gssapi.Negotiate(log)

http_server = server.HTTPServer
try:
    from http.server import ThreadingHTTPServer
    http_server = server.ThreadingHTTPServer
except ImportError:
    pass

VERSION = "2.0"

def _check(request, function):
    """Check if user has sufficient rights for performing ``request``."""
    # ``_check`` decorator can access ``request`` protected functions
    # pylint: disable=W0212
    owner = user = password = None
    negotiate_success = False

    if request._collection:
        owner = request._collection.owner

    authorization = request.headers.get("Authorization", None)
    if authorization:
        if authorization.startswith("Basic"):
            challenge = authorization.lstrip("Basic").strip().encode("ascii")
            plain = request._decode(base64.b64decode(challenge))
            user, password = plain.split(":")
        elif negotiate.enabled():
            user, negotiate_success = negotiate.try_aaa(authorization, request, owner)

    # Also send UNAUTHORIZED if there's no collection. Otherwise one
    # could probe the server for (non-)existing collections.
    if request.server.acl.has_right(owner, user, password) or negotiate_success:
        function(request, context={"user": user, "user-agent": request.headers.get("User-Agent", None)})
    else:
        request.send_calypso_response(client.UNAUTHORIZED, 0)
        if negotiate.enabled():
            request.send_header("WWW-Authenticate", "Negotiate")
        request.send_header(
            "WWW-Authenticate",
            'Basic realm="Calypso CalDAV/CardDAV server - password required"')
        request.end_headers()
    # pylint: enable=W0212


class HTTPServer(server.HTTPServer):
    """HTTP server."""
    PROTOCOL = "http"

    # Maybe a Pylint bug, ``__init__`` calls ``server.HTTPServer.__init__``
    # pylint: disable=W0231
    def __init__(self, address, handler):
        """Create server."""
        http_server.__init__(self, address, handler)
        self.acl = acl.load()
    # pylint: enable=W0231


class HTTPSServer(HTTPServer):
    """HTTPS server."""
    PROTOCOL = "https"

    def server_bind(self):
        HTTPServer.server_bind(self)

        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        context.load_cert_chain(os.path.expanduser(config.get("server", "certificate")),
                                keyfile=os.path.expanduser(config.get("server", "key")))

        self.socket = context.wrap_socket(self.socket,server_side=True)

class CollectionHTTPHandler(server.BaseHTTPRequestHandler):
    """HTTP requests handler for WebDAV collections."""
    _encoding = config.get("encoding", "request")

    # Decorator checking rights before performing request
    check_rights = lambda function: lambda request: _check(request, function)

    # We do set Content-Length on all replies, so we can use HTTP/1.1
    # with multiple requests (as desired by the android CalDAV sync program

    protocol_version = 'HTTP/1.1'

    timeout = 90

    server_version = "Calypso/%s" % VERSION
    queued_headers = {}

    def queue_header(self, keyword, value):
        self.queued_headers[keyword] = value

    def end_headers(self):
        """
        Send out all queued headers and invoke or super classes
        end_header.
        """
        if self.queued_headers:
            for keyword, val in self.queued_headers.items():
                self.send_header(keyword, val)
            self.queued_headers = {}
        return server.BaseHTTPRequestHandler.end_headers(self)

    def address_string(self):
        return str(self.client_address[0])

    def send_connection_header(self):
        conntype = "Close"
        if self.close_connection == 0:
            conntype = "Keep-Alive"
        self.send_header("Connection", conntype)

    def send_calypso_response(self, response, length):
        self.send_response(response)
        self.send_connection_header()
        self.send_header("Content-Length", length)
        for header, value in config.items('headers'):
            self.send_header(header, value)


    def handle_one_request(self):
        """Handle a single HTTP request.

        You normally don't need to override this method; see the class
        __doc__ string for information on how to handle specific HTTP
        commands such as GET and POST.

        """
        try:
            self.wfile.flush()
            self.close_connection = 1

            self.connection.settimeout(5)

            self.raw_requestline = self.rfile.readline(65537)

            self.connection.settimeout(90)

            if len(self.raw_requestline) > 65536:
                log.error("Read request too long")
                self.requestline = ''
                self.request_version = ''
                self.command = ''
                self.send_error(414)
                return
            if not self.raw_requestline:
                log.error("Connection closed")
                return
            log.debug("First line '%s'", self.raw_requestline)
            if not self.parse_request():
                # An error code has been sent, just exit
                self.close_connection = 1
                return
            # parse_request clears close_connection on all http/1.1 links
            # it should only do this if a keep-alive header is seen
            self.close_connection = 1
            conntype = self.headers.get('Connection', "")
            if (conntype.lower() == 'keep-alive'
                and self.protocol_version >= "HTTP/1.1"):
                log.debug("keep-alive")
                self.close_connection = 0
            reqlen = self.headers.get('Content-Length',"0")
            log.debug("reqlen %s", reqlen)
            self.xml_request = self.rfile.read(int(reqlen))
            mname = 'do_' + self.command
            if not hasattr(self, mname):
                log.error("Unsupported method (%r)", self.command)
                self.send_error(501, "Unsupported method (%r)" % self.command)
                return
            method = getattr(self, mname)
            method()
            self.wfile.flush() #actually send the response if not already done.
        except socket.timeout as e:
            #a read or a write timed out.  Discard this connection
            log.error("Request timed out: %r", e)
            self.close_connection = 1
            return
        except ssl.SSLError as x:
            #an io error. Discard this connection
            log.error("SSL request error: %r", x.args[0])
            self.close_connection = 1
            return


    collections = {}
    collections_lock = threading.Lock()

    @property
    def _collection(self):
        """The ``webdav.Collection`` object corresponding to the given path."""
        path = paths.collection_from_path(self.path)
        if not path:
            return None
        with CollectionHTTPHandler.collections_lock:
            if not path in CollectionHTTPHandler.collections:
                CollectionHTTPHandler.collections[path] = webdav.Collection(path)
            return CollectionHTTPHandler.collections[path]

    def _decode(self, text):
        """Try to decode text according to various parameters."""
        # List of charsets to try
        charsets = []

        # First append content charset given in the request
        content_type = self.headers.get("Content-Type", None)
        if content_type and "charset=" in content_type:
            charsets.append(content_type.split("charset=")[1].strip())
        # Then append default Calypso charset
        charsets.append(self._encoding)
        # Then append various fallbacks
        charsets.append("utf-8")
        charsets.append("iso8859-1")

        # Try to decode
        for charset in charsets:
            try:
                return text.decode(charset)
            except UnicodeDecodeError:
                pass
        raise UnicodeDecodeError

    # Naming methods ``do_*`` is OK here
    # pylint: disable=C0103

    @check_rights
    def do_GET(self, context):
        """Manage GET request."""
        self.do_get_head(context, True)

    @check_rights
    def do_HEAD(self, context):
        """Manage HEAD request."""
        self.do_get_head(context, False)

    def do_get_head(self, context, is_get):
        """Manage either GET or HEAD request."""

        self._answer = ''
        answer_text = ''
        code = client.OK
        last_modified = ''
        try:
            item_name = paths.resource_from_path(self.path)
            if self._collection:
                with self._collection:
                    if item_name:
                        # Get collection item
                        item = self._collection.get_item(item_name)
                        if item:
                            if is_get:
                                answer_text = item.text
                            etag = item.etag
                        else:
                            code = client.GONE
                    else:
                        # Get whole collection
                        if is_get:
                            answer_text = self._collection.text
                        etag = self._collection.etag
                    last_modified = self._collection.last_modified

                if len(answer_text):
                    try:
                        self._answer = answer_text.encode(self._encoding,"xmlcharrefreplace")
                    except UnicodeDecodeError:
                        answer_text = answer_text.decode(errors="ignore")
                        self._answer = answer_text.encode(self._encoding,"ignore")
            else:
                code = client.NOT_FOUND

        except Exception:
            log.exception("Failed HEAD for %s", self.path)
            code = client.BAD_REQUEST

        self.send_calypso_response(code, len(self._answer))
        if code == client.OK:
            self.send_header("Content-Type", "text/calendar")
            self.send_header("Last-Modified", email.utils.formatdate(time.mktime(last_modified)))
            self.send_header("ETag", etag)
        self.end_headers()
        if len(self._answer):
            self.wfile.write(self._answer)

    def if_match(self, items):
        for item in items:
            header = self.headers.get("If-Match", item.etag)
            header = email.utils.unquote(header)
            log.debug("header '%s' etag '%s'" % (header, item.etag))
            if header == item.etag:
                return True
            quoted = '"' + item.etag + '"'
            if header == quoted:
                return True
            extraquoted = email.utils.quote(quoted)
            if header == extraquoted:
                return True
        return False

    @check_rights
    def do_DELETE(self, context):
        """Manage DELETE request."""
        self._answer = ''
        code = client.NO_CONTENT
        try:
            item_name = paths.resource_from_path(self.path)
            with self._collection:
                items = self._collection.get_items(item_name)

                if len(items) and self.if_match(items):
                    # No ETag precondition or precondition verified, delete item
                    self._answer = xmlutils.delete(self.path, self._collection, context=context)

                elif not len(items):
                    # Item does not exist
                    code = client.NOT_FOUND
                else:
                    # No item or ETag precondition not verified, do not delete item
                    code = client.PRECONDITION_FAILED

        except Exception:
            log.exception("Failed DELETE for %s", self.path)
            code = client.BAD_REQUEST

        self.send_calypso_response(code, len(self._answer))
        if len(self._answer):
            self.send_header("Content-Type", "text/xml")
        self.end_headers()
        if len(self._answer):
            self.wfile.write(self._answer)

    @check_rights
    def do_MKCALENDAR(self, context):
        """Manage MKCALENDAR request."""
        self.send_calypso_response(client.CREATED, 0)
        self.end_headers()

    def do_OPTIONS(self):
        """Manage OPTIONS request."""
        self.send_calypso_response(client.OK, 0)
        self.send_header(
            "Allow", "DELETE, HEAD, GET, MKCALENDAR, "
            "OPTIONS, PROPFIND, PUT, REPORT")
        self.send_header("DAV", "1, access-control, calendar-access, addressbook")
        self.end_headers()

    @check_rights
    def do_PROPFIND(self, context):
        """Manage PROPFIND request."""
        try:
            xml_request = self.xml_request
            log.debug("PROPFIND %s", xml_request)
            depth = self.headers.get("depth", "infinity")
            if depth != "infinity":
                with self._collection:
                    self._answer = xmlutils.propfind(
                        self.path, xml_request, self._collection,
                        depth, context)
                status = client.MULTI_STATUS
            else:
                self._answer = xmlutils.propfind_deny()
                status = client.FORBIDDEN

            if len(self._answer) < 100:
                log.debug("PROPFIND ANSWER %s", self._answer)
            else:
                log.debug("PROPFIND ANSWER len %d", len(self._answer))

            self.send_calypso_response(status, len(self._answer))
            self.send_header("DAV", "1, calendar-access")
            self.send_header("Content-Type", "text/xml")
            self.end_headers()
            self.wfile.write(self._answer)
        except Exception:
            log.exception("Failed PROPFIND for %s", self.path)
            self.send_calypso_response(client.BAD_REQUEST, 0)
            self.end_headers()

    @check_rights
    def do_SEARCH(self, context):
        """Manage SEARCH request."""
        try:
            self.send_calypso_response(client.NO_CONTENT, 0)
            self.end_headers()
        except Exception:
            log.exception("Failed SEARCH for %s", self.path)
            self.send_calypso_response(client.BAD_REQUEST, 0)
            self.end_headers()

    @check_rights
    def do_PUT(self, context):
        """Manage PUT request."""

        code = client.CREATED
        etag = None
        try:
            item_name = paths.resource_from_path(self.path)
            with self._collection:
                items = self._collection.get_items(item_name)
                if not len(items) or self.if_match(items):

                    # PUT allowed in 3 cases
                    # Case 1: No item and no ETag precondition: Add new item
                    # Case 2: Item and ETag precondition verified: Modify item
                    # Case 3: Item and no Etag precondition: Force modifying item
                    webdav_request = self._decode(self.xml_request)
                    new_item = xmlutils.put(self.path, webdav_request, self._collection, context=context)

                    log.debug("item_name %s new_name %s", item_name, new_item.name)
                    etag = new_item.etag
                    #log.debug("replacement etag %s", etag)

                    code = client.CREATED
                    etag = new_item.etag
                else:
                    #log.debug("Precondition failed")
                    # PUT rejected in all other cases
                    code = client.PRECONDITION_FAILED

        except Exception:
            log.exception('Failed PUT for %s', self.path)
            code = client.BAD_REQUEST

        self.send_calypso_response(code, 0)
        if etag:
            self.send_header('ETag', etag)
        self.end_headers()

    @check_rights
    def do_REPORT(self, context):
        """Manage REPORT request."""
        try:
            xml_request = self.xml_request
            log.debug("REPORT %s %s", self.path, xml_request)
            with self._collection:
                self._answer = xmlutils.report(self.path, xml_request, self._collection)

            if len(self._answer) < 100:
                log.debug("REPORT ANSWER %s" % self._answer)
            else:
                log.debug("REPORT ANSWER %d" % len(self._answer))
            self.send_calypso_response(client.MULTI_STATUS, len(self._answer))
            self.send_header("Content-Type", "text/xml")
            self.end_headers()
            self.wfile.write(self._answer)
        except Exception:
            log.exception("Failed REPORT for %s", self.path)
            self.send_calypso_response(client.BAD_REQUEST, 0)
            self.end_headers()

    # pylint: enable=C0103



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This file is part of Calypso - CalDAV/CardDAV/WebDAV Server
# Copyright © 2011 Keith Packard
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

# This file is just a script, allow [a-z0-9]* variable names
# pylint: disable-msg=C0103

# ``import calypso`` refers to the ``calypso`` module, not ``calypso.py``
# pylint: disable-msg=W0406

"""
Calypso Server entry point.

Launch the Calypso Server according to configuration and command-line
arguments.

"""

# TODO: Manage smart and configurable logs

import daemon
import lockfile
import logging
import optparse
import os
import sys

import calypso
import calypso.webdav as webdav

# Get command-line options
parser = optparse.OptionParser(version=calypso.VERSION)
parser.add_option(
    "-d", "--daemon", action="store_true",
    default=calypso.config.getboolean("server", "daemon"),
    help="launch as daemon")
parser.add_option(
    "-f", "--foreground", action="store_false", dest="daemon",
    help="launch in foreground (opposite of --daemon)")
parser.add_option(
    "-H", "--host",
    default=calypso.config.get("server", "host"),
    help="set server hostname")
parser.add_option(
    "-p", "--port", type="int",
    default=calypso.config.getint("server", "port"),
    help="set server port")
parser.add_option(
    "-s", "--ssl", action="store_true",
    default=calypso.config.getboolean("server", "ssl"),
    help="use SSL connection")
parser.add_option(
    "-S", "--no-ssl", action="store_false", dest="ssl",
    help="do not use SSL connection (opposite of --ssl)")
parser.add_option(
    "-k", "--key",
    default=calypso.config.get("server", "key"),
    help="private key file ")
parser.add_option(
    "-c", "--certificate",
    default=calypso.config.get("server", "certificate"),
    help="certificate file ")
parser.add_option(
    "-i", "--import", dest="import_dest")
parser.add_option(
    "-g", "--debug", action="store_true",
    default=False,
    help="enable debug logging")
parser.add_option(
    "-P", "--pid-file", dest="pidfile",
    default=calypso.config.get("server", "pidfile"),
    help="set location of process-id file")

(options, args) = parser.parse_args()

# Update Calypso configuration according to options
for option in parser.option_list:
    key = option.dest
    if key:
        value = getattr(options, key)
        calypso.config.set("server", key, value)

log = logging.getLogger()
ch = logging.StreamHandler()

# Handle debugging option and log levels
if options.debug:
    log.setLevel(logging.DEBUG)
    ch.setLevel(logging.DEBUG)
    logging.basicConfig(level=logging.DEBUG)
    log.debug("enable debugging")
else:
    log.setLevel(logging.WARN)
    ch.setLevel(logging.WARN)
    logging.basicConfig(level=logging.WARN)


# Run import if requested
if options.import_dest:
    try:
        collection = webdav.Collection(options.import_dest)
    except Exception:
        log.error("Cannot open collection %s" % options.import_dest)
        sys.exit(1)
    success = True
    for arg in args:
        if not collection.import_file(arg):
            success = False
    if success:
        sys.exit(0)
    else:
        sys.exit(1)

def run_server():
    try:
        # Launch server
        log.debug("Starting HTTP%s server on %s:%d" % ("S" if options.ssl else "",
                                                       options.host if options.host else "*",
                                                       options.port))
        server_class = calypso.HTTPSServer if options.ssl else calypso.HTTPServer
        server = server_class(
            (options.host, options.port), calypso.CollectionHTTPHandler)
        server.serve_forever(poll_interval=10)
    except KeyboardInterrupt:
        server.socket.close()

# If foreground execution is requested, just run the server
if not options.daemon:
    run_server()
    sys.exit(0)

# Otherwise, daemonize Calypso
context = daemon.DaemonContext()
context.umask = 0o002
if options.pidfile:
    from lockfile import pidlockfile
    # Generate a pidfile where requested
    context.pidfile = pidlockfile.PIDLockFile(options.pidfile)
with context:
    run_server()

# vim: set ts=4 sw=4 et si :



# -*- coding: utf-8 -*-
#
# This file is part of Calypso Server - Calendar Server
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Calypso configuration module.

Give a configparser-like interface to read and write configuration.

"""

import os
import sys
# Manage Python2/3 different modules
# pylint: disable=F0401
try:
    from configparser import RawConfigParser as ConfigParser
except ImportError:
    from ConfigParser import RawConfigParser as ConfigParser
# pylint: enable=F0401


# Default configuration
INITIAL_CONFIG = {
    "server": {
        "host": "",
        "port": "5233",
        "daemon": "False",
        "ssl": "False",
        "certificate": "/etc/apache2/ssl/server.crt",
        "key": "/etc/apache2/ssl/server.key",
        "pidfile": "/var/run/calypso.pid",
        "user_principal": "/%(user)s",
        "base_prefix": "/",
    },
    "encoding": {
        "request": "utf-8",
        "stock": "utf-8"},
    "acl": {
        "type": "fake",
        "personal": "False",
        "filename": "/etc/calypso/users",
        "encryption": "crypt",
        "pam_service": "passwd",
    },
    "storage": {
        "folder": os.path.expanduser("~/.config/calypso/calendars")
    },
    "headers": {
    },
}

# Create a ConfigParser and configure it
_CONFIG_PARSER = ConfigParser()

for section, values in INITIAL_CONFIG.items():
    _CONFIG_PARSER.add_section(section)
    for key, value in values.items():
        _CONFIG_PARSER.set(section, key, value)

_CONFIG_PARSER.read("/etc/calypso/config")
_CONFIG_PARSER.read(os.path.expanduser("~/.config/calypso/config"))
cfg = os.getenv("CALYPSO_CONFIG")
if cfg:
    _CONFIG_PARSER.read(cfg)

# Wrap config module into ConfigParser instance
sys.modules[__name__] = _CONFIG_PARSER



# -*- coding: utf-8 -*-
#
# This file is part of Calypso Server - Calendar Server
# Copyright © 2013 Joseph Nahmias
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
PAM authentication

Use Pluggable Authentication Modules (PAM) system on Linux
for checking users/passwords.

"""

import logging
import PAM

from calypso import config

LOG = logging.getLogger()
SVC = config.get("acl", "pam_service")
PERSONAL = config.getboolean("acl", "personal")

def has_right(owner, user, password):
    """Check if ``user``/``password`` couple is valid."""
    LOG.debug("owner %s user %s", owner, user)
    if owner and owner != user and PERSONAL:
        return False
    def pam_conv(auth, query_list, userData):
        result = []
        result.append((password, 0))
        return result
    try:
        auth = PAM.pam()
        auth.start(SVC)
        auth.set_item(PAM.PAM_USER, user)
        auth.set_item(PAM.PAM_CONV, pam_conv)
        auth.authenticate()
        auth.acct_mgmt()
        return True
    except PAM.error as resp:
        LOG.debug('PAM error: %s', resp)
    return False

# vi: set ts=4 sw=4 et si :



# vim: set fileencoding=utf-8 :
"""Test matching filter handling """

import subprocess
import tempfile
import shutil
import unittest
import xml.etree.ElementTree as ET

from calypso.webdav import Collection
from calypso import xmlutils

from .testutils import CalypsoTestCase


class TestMatchFilterElement(CalypsoTestCase):
    test_vcal = "tests/data/import.vcalendar"

    def test_start_end(self):
        """
Check that the time-range parser accept ranges where start or stop is
missing.
"""
        valid_xml_request1 ="""
<calendar-query xmlns="urn:ietf:params:xml:ns:caldav">
 <prop xmlns="DAV:">
  <getetag xmlns="DAV:"/>
  <resourcetype xmlns="DAV:"/>
 </prop>
 <filter xmlns="urn:ietf:params:xml:ns:caldav">
  <comp-filter xmlns="urn:ietf:params:xml:ns:caldav" name="VCALENDAR">
    <time-range xmlns="urn:ietf:params:xml:ns:caldav" start="20151021T201004Z"/>
  </comp-filter>
 </filter>
</calendar-query>
"""
        valid_xml_request2 ="""
<calendar-query xmlns="urn:ietf:params:xml:ns:caldav">
 <prop xmlns="DAV:">
  <getetag xmlns="DAV:"/>
  <resourcetype xmlns="DAV:"/>
 </prop>
 <filter xmlns="urn:ietf:params:xml:ns:caldav">
  <comp-filter xmlns="urn:ietf:params:xml:ns:caldav" name="VCALENDAR">
    <time-range xmlns="urn:ietf:params:xml:ns:caldav" end="20151021T201004Z"/>
  </comp-filter>
 </filter>
</calendar-query>
"""
        invalid_xml_request1 ="""
<calendar-query xmlns="urn:ietf:params:xml:ns:caldav">
 <prop xmlns="DAV:">
  <getetag xmlns="DAV:"/>
  <resourcetype xmlns="DAV:"/>
 </prop>
 <filter xmlns="urn:ietf:params:xml:ns:caldav">
  <comp-filter xmlns="urn:ietf:params:xml:ns:caldav" name="VCALENDAR">
    <time-range xmlns="urn:ietf:params:xml:ns:caldav"/>
  </comp-filter>
 </filter>
</calendar-query>
"""
        collection = Collection("")
        self.assertTrue(collection.import_file(self.test_vcal))
        self.assertEqual(len(collection.items), 1)

        # Tried calling do_REPORT() directly, but lacked the arguments
        # needed to get the CollectionHTTPHandler class working.  Use
        # match_filter() directly instead.
        truecount = 0
        for xml_request in [valid_xml_request1, valid_xml_request2]:
            root = ET.fromstring(xml_request)
            filter_element = root.find(xmlutils._tag("C", "filter"))
            for item in collection.items:
                answer = xmlutils.match_filter(item, filter_element)
                if answer:
                    truecount += 1
        # The text vcalendar entry is either before or after the cutoff point.
        self.assertEqual(truecount, 1)

        for xml_request in [invalid_xml_request1]:
            root = ET.fromstring(xml_request)
            filter_element = root.find(xmlutils._tag("C", "filter"))
            for item in collection.items:
                with self.assertRaisesRegex(ValueError, "time-range missing both start and stop attribute"):
                    xmlutils.match_filter(item, filter_element)
        # The text vcalendar entry is either before or after the cutoff point.




# -*- coding: utf-8 -*-
#
# This file is part of Calypso Server - Calendar Server
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Fake ACL.

No rights management.

"""

def has_right(*args, **kwargs):
    """Check if ``user``/``password`` couple is valid."""
    return True



# vim: set fileencoding=utf-8 :

import subprocess
import tempfile
import shutil
import unittest

from calypso.webdav import Collection
from calypso import paths

from .testutils import CalypsoTestCase


class TestCollection(CalypsoTestCase):
    test_vcard = "tests/data/import.vcard"
    test_resource_with_slash = "tests/data/from-tripsync.ics"

    def test_import_file(self):
        collection = Collection("")
        self.assertTrue(collection.import_file(self.test_vcard))
        self.assertEqual(len(collection.items), 2)
        org = u'Universitetet i Tromsø'
        self.assertEqual(org, collection.items[0].object.org.value[0])

    def test_remove_existent_item(self):
        collection = Collection("")
        self.assertTrue(collection.import_file(self.test_vcard))
        self.assertEqual(len(collection.items), 2)
        name = collection.items[0].name
        collection.remove(name, {})
        self.assertEqual(len(collection.items), 1)

    def test_remove_nonexistent_item(self):
        collection = Collection("")
        self.assertTrue(collection.import_file(self.test_vcard))
        self.assertEqual(len(collection.items), 2)
        collection.remove("doesnotexist", {})
        self.assertEqual(len(collection.items), 2)

    def test_uid_with_slash(self):
        collection = Collection("/")
        self.assertTrue(collection.import_file(self.test_resource_with_slash))
        self.assertEqual(len(collection.items), 1)
        veventuid = collection.items[0].object.vevent.uid.value
        r = paths.resource_from_path(veventuid)
        c = paths.collection_from_path(veventuid)
        self.assertEqual(r, veventuid)
        self.assertEqual("/", c)



import shutil
import subprocess
import tempfile
import unittest

import calypso.config


class CalypsoTestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        calypso.config.set('storage', 'folder', self.tmpdir)
        subprocess.call(["git", "init", self.tmpdir])
        subprocess.call(["git", "config", "user.email", "calypso@example.com"], cwd=self.tmpdir)
        subprocess.call(["git", "config", "user.name", "cal Ypso"], cwd=self.tmpdir)

    def tearDown(self):
        if self.tmpdir:
            shutil.rmtree(self.tmpdir)



# -*- coding: utf-8 -*-
#
# This file is part of Calypso Server - Calendar Server
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Htpasswd ACL.

Load the list of login/password couples according a the configuration file
created by Apache ``htpasswd`` command. Plain-text, crypt and sha1 are
supported, but md5 is not (see ``htpasswd`` man page to understand why).

"""

import base64
import hashlib
import os.path
import logging

from calypso import config

log = logging.getLogger()

def _plain(hash_value, password):
    """Check if ``hash_value`` and ``password`` match using plain method."""
    return hash_value == password


def _crypt(hash_value, password):
    """Check if ``hash_value`` and ``password`` match using crypt method."""
    # The ``crypt`` module is only present on Unix, import if needed
    import crypt
    return crypt.crypt(password, hash_value) == hash_value


def _sha1(hash_value, password):
    """Check if ``hash_value`` and ``password`` match using sha1 method."""
    hash_value = hash_value.replace("{SHA}", "").encode("ascii")
    password = password.encode(config.get("encoding", "stock"))
    sha1 = hashlib.sha1() # pylint: disable=E1101
    sha1.update(password)
    return sha1.digest() == base64.b64decode(hash_value)


def has_right(owner, user, password):
    """Check if ``user``/``password`` couple is valid."""
    log.debug("owner '%s' user '%s'", owner, user)
    for line in open(FILENAME).readlines():
        if line.strip():
            login, hash_value = line.strip().split(":", 1)
            if login == user and (not PERSONAL or user == owner):
                return CHECK_PASSWORD(hash_value, password)
    return False


FILENAME = os.path.expanduser(config.get("acl", "filename"))
PERSONAL = config.getboolean("acl", "personal")
CHECK_PASSWORD = locals()["_%s" % config.get("acl", "encryption")]



# -*- coding: utf-8 -*-
#
# This file is part of Calypso - CalDAV/CardDAV/WebDAV Server
# Copyright © 2011 Keith Packard
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

import urllib
import urllib.request
import os.path
import posixpath # the semantics of urls follow posix rules, not platform dependent rules
import logging

from . import config

log = logging.getLogger()

#
# Various URL manipulation routines
#

# The first element of the URL is always the object owner

def url_to_owner(path):
    return path.strip("/").split("/")[0]

#
# Return the folder under which all data is stored
#

def data_root():
    return os.path.expanduser(config.get("storage", "folder")).rstrip('/')


#
# Return the base path for the web server.
#

def base_prefix():
    return config.get("server", "base_prefix").rstrip("/")

#
# Given a URL, convert it to an absolute path name by
# prepending the storage folder name
#
# Results are guaranteed to reside under data_root().
#

def url_to_file(url):
    if url.startswith(base_prefix()):
        url = "/" + url[len(base_prefix()):].lstrip("/")
    tail = urllib.request.url2pathname(url.strip("/"))
    # eliminate .. components, and potential double leading slashes
    tail = posixpath.normpath('/' + tail).lstrip('/')
    file = os.path.join(data_root(), tail)
    return file


#
# Does the provided URL reference a collection? This
# is done by seeing if the resulting path is a directory
#

def is_collection(url):
    urlpath = url_to_file(url)
    return os.path.isdir(urlpath)

#
# Given a URL, return the parent URL by stripping off
# the last path element
#

def parent_url(path):
    path_parts = path.strip("/").split("/")
    path_parent = path_parts[0:len(path_parts)-1]
    new_path = "/" + "/".join(path_parent)
    return new_path

#
# Given a URL, return the child URL, which is
# the last path element
#

def child_url(path):
    path_parts = path.strip("/").split("/")
    return path_parts[len(path_parts)-1]

#
# If the given URL references a resource, then
# return the name of that resource. Otherwise,
# return None
#

def resource_from_path(path):
    """Return Calypso item name from ``path``."""

    child_path = None
    collection = path

    while collection != '/' and not is_collection(collection):
        child = child_url(collection)
        if child_path:
            child_path = child + "/" + child_path
        else:
            child_path = child
        collection = parent_url(collection)

    if child_path:
        name = urllib.parse.unquote(child_path)
    else:
        name = None

    return name

#
# Return the collection name for the given URL. That's
# just the URL if it refers to a collection itself, otherwise
# it's the parent of the provided URL
#

def collection_from_path(path):
    """Returns Calypso collection name from ``path``."""

    collection = path
    while collection != '/' and not is_collection(collection):
        collection = parent_url(collection)

    if not collection:
        log.debug("No collection found for path %s", path)
        return None

    # unquote, strip off any trailing slash, then clean up /../ and // entries
    collection = "/" + urllib.parse.unquote(collection).strip("/")

    return collection




# -*- coding: utf-8 -*-
#
# This file is part of Calypso - CalDAV/CardDAV/WebDAV Server
# Copyright © 2011 Keith Packard
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
XML and iCal requests manager.

Note that all these functions need to receive unicode objects for full
iCal requests (PUT) and string objects with charset correctly defined
in them for XML requests (all but PUT).

"""

import xml.etree.ElementTree as ET
import time
import dateutil
import dateutil.parser
import dateutil.rrule
import dateutil.tz
import datetime
import email.utils
import logging

from . import client, config, webdav, paths

__package__ = 'calypso.xmlutils'

NAMESPACES = {
    "C": "urn:ietf:params:xml:ns:caldav",
    "A": "urn:ietf:params:xml:ns:carddav",
    "D": "DAV:",
    "E": "http://apple.com/ns/ical/",
    "CS": "http://calendarserver.org/ns/"}

log = logging.getLogger(__name__)

def _tag(short_name, local):
    """Get XML Clark notation {uri(``short_name``)}``local``."""
    return "{%s}%s" % (NAMESPACES[short_name], local)


def _response(code):
    """Return full W3C names from HTTP status codes."""
    return "HTTP/1.1 %i %s" % (code, client.responses[code])

def delete(path, collection, context):
    """Read and answer DELETE requests.

    Read rfc4918-9.6 for info.

    """
    # Reading request
    collection.remove(paths.resource_from_path(path), context=context)

    # Writing answer
    multistatus = ET.Element(_tag("D", "multistatus"))
    response = ET.Element(_tag("D", "response"))
    multistatus.append(response)

    href = ET.Element(_tag("D", "href"))
    href.text = path
    response.append(href)

    status = ET.Element(_tag("D", "status"))
    status.text = _response(200)
    response.append(status)

    return ET.tostring(multistatus, config.get("encoding", "request"))


def propfind(path, xml_request, collection, depth, context):
    """Read and answer PROPFIND requests.

    Read rfc4918-9.1 for info.

    """

    item_name = paths.resource_from_path(path)
    collection_name = paths.collection_from_path(path)

    if xml_request:
        # Reading request
        root = ET.fromstring(xml_request)

        prop_element = root.find(_tag("D", "prop"))
    else:
        prop_element = None

    if prop_element is not None:
        props = [prop.tag for prop in prop_element]
    else:
        props = [_tag("D", "resourcetype"),
                 _tag("D", "owner"),
                 _tag("D", "getcontenttype"),
                 _tag("D", "getetag"),
                 _tag("D", "principal-collection-set"),
                 _tag("C", "supported-calendar-component-set"),
                 _tag("D", "supported-report-set"),
                 _tag("D", "current-user-privilege-set"),
                 _tag("D", "getcontentlength"),
                 _tag("D", "getlastmodified")]


    # Writing answer
    multistatus = ET.Element(_tag("D", "multistatus"))

    if collection:
        if item_name:
            item = collection.get_item(item_name)
            log.debug("item_name %s item %s" % (item_name, item))
            if item:
                items = [item]
            else:
                items = []
        else:
            if depth == "0":
                items = [collection]
            else:
                # We limit ourselves to depth == 1
                items = [collection] + collection.items + collection.dirs
    else:
        items = []

    for item in items:
        is_collection = item.is_collection

        response = ET.Element(_tag("D", "response"))
        multistatus.append(response)

        href = ET.Element(_tag("D", "href"))
        href.text = item.urlpath
        response.append(href)

        propstat = ET.Element(_tag("D", "propstat"))
        response.append(propstat)

        prop = ET.Element(_tag("D", "prop"))
        propstat.append(prop)

        for tag in props:
            element = ET.Element(tag)
            if tag == _tag("D", "resourcetype") and is_collection:
                if item.is_calendar:
                    tag = ET.Element(_tag("C", "calendar"))
                    element.append(tag)
                if item.is_addressbook:
                    tag = ET.Element(_tag("A", "addressbook"))
                    element.append(tag)
                tag = ET.Element(_tag("D", "collection"))
                element.append(tag)
            elif tag == _tag("D", "owner"):
                element.text = collection.owner
            elif tag == _tag("D", "getcontenttype"):
                if item.tag == 'VCARD':
                    element.text = "text/vcard"
                else:
                    element.text = "text/calendar"
            elif tag == _tag("CS", "getctag") and is_collection:
                element.text = item.ctag
            elif tag == _tag("D", "getetag"):
                element.text = item.etag
            elif tag == _tag("D", "displayname") and is_collection:
                element.text = item.name
            elif tag == _tag("E", "calendar-color") and is_collection:
                element.text = item.color
                if not element.text:
                    element.text = collection.color
            elif tag == _tag("D", "principal-URL"):
                # TODO: use a real principal URL, read rfc3744-4.2 for info
                tag = ET.Element(_tag("D", "href"))
                tag.text = path
                element.append(tag)
            elif tag in (
                _tag("D", "principal-collection-set"),
                _tag("C", "calendar-user-address-set"),
                _tag("C", "calendar-home-set"),
                _tag("A", "addressbook-home-set")):
                tag = ET.Element(_tag("D", "href"))
                tag.text = path
                element.append(tag)
            elif tag == _tag("C", "supported-calendar-component-set"):
                comp = ET.Element(_tag("C", "comp"))
                comp.set("name", "VTODO") # pylint: disable=W0511
                element.append(comp)
                comp = ET.Element(_tag("C", "comp"))
                comp.set("name", "VEVENT")
                element.append(comp)
            elif tag == _tag("D", "supported-report-set"):
                tag = ET.Element(_tag("C", "calendar-multiget"))
                element.append(tag)
                tag = ET.Element(_tag("C", "filter"))
                element.append(tag)
            elif tag == _tag("D", "current-user-privilege-set"):
                privilege = ET.Element(_tag("D", "privilege"))
                privilege.append(ET.Element(_tag("D", "all")))
                element.append(privilege)
            elif tag == _tag("D", "getcontentlength"):
                element.text = item.length
            elif tag == _tag("D", "getlastmodified"):
#                element.text = time.strftime("%a, %d %b %Y %H:%M:%S +0000", item.last_modified)
#                element.text = email.utils.formatdate(item.last_modified)
                element.text = email.utils.formatdate(time.mktime(item.last_modified))
            elif tag == _tag("D", "current-user-principal"):
                tag = ET.Element(_tag("D", "href"))
                tag.text = config.get("server", "user_principal") % context
                element.append(tag)
            elif tag in (_tag("A", "addressbook-description"),
                         _tag("C", "calendar-description")) and is_collection:
                element.text = item.get_description()
                if not element.text:
                    element.text = "no description"
            prop.append(element)

        status = ET.Element(_tag("D", "status"))
        status.text = _response(200)
        propstat.append(status)

    return ET.tostring(multistatus, config.get("encoding", "request"))


def propfind_deny():
    """Answer an infinity PROPFIND requests.

    Read rfc4918-9.1.1 for info.
    """
    error = ET.Element(_tag("D", "error"))
    prec_code = ET.Element(_tag("D", "propfind-finite-depth"))
    error.append(prec_code)
    return ET.tostring(error, config.get("encoding", "request"))


def put(path, webdav_request, collection, context):
    """Read PUT requests."""
    name = paths.resource_from_path(path)
    log.debug('xmlutils put path %s name %s', path, name)
    old_item = collection.get_item(name)
    if old_item:
        # PUT is modifying an existing item
        log.debug('Replacing item named %s', name)
        return collection.replace(name, webdav_request, context=context)
    else:
        # PUT is adding a new item
        log.debug('Putting a new item, because name %s is not known', name)
        return collection.append(name, webdav_request, context=context)


def match_filter_element(vobject, fe):
    if fe.tag == _tag("C", "comp-filter"):
        comp = fe.get("name")
        if comp:
            if comp == vobject.name:
                hassub = False
                submatch = False
                for fc in fe:
                    if match_filter_element(vobject, fc):
                        submatch = True
                        break
                    for vc in vobject.getChildren():
                        hassub = True
                        if match_filter_element (vc, fc):
                            submatch = True
                            break
                    if submatch:
                        break
                if not hassub or submatch:
                    return True
        return False
    elif fe.tag == _tag("C", "time-range"):
        try:
            rruleset = vobject.rruleset
        except AttributeError:
            return False
        start = fe.get("start")
        end = fe.get("end")
        # According to RFC 4791, one of start and stop must be set,
        # but the other can be empty.  If both are empty, the
        # specification is violated.
        if start is None and end is None:
            msg = "time-range missing both start and stop attribute (required by RFC 4791)"
            log.error(msg)
            raise ValueError(msg)
        # RFC 4791 state if start is missing, assume it is -infinity
        if start is None:
            start = "00010101T000000Z"  # start of year one
        # RFC 4791 state if end is missing, assume it is +infinity
        if end is None:
            end = "99991231T235959Z"  # last date with four digit year
        if rruleset is None:
            rruleset = dateutil.rrule.rruleset()
            dtstart = vobject.dtstart.value
            try:
                dtstart = datetime.datetime.combine(dtstart, datetime.time())
            except Exception:
                0
            if dtstart.tzinfo is None:
                dtstart = dtstart.replace(tzinfo = dateutil.tz.tzlocal())
            rruleset.rdate(dtstart)
        start_datetime = dateutil.parser.parse(start)
        if start_datetime.tzinfo is None:
            start_datetime = start_datetime.replace(tzinfo = dateutil.tz.tzlocal())
        end_datetime = dateutil.parser.parse(end)
        if end_datetime.tzinfo is None:
            end_datetime = end_datetime.replace(tzinfo = dateutil.tz.tzlocal())
        try:
            if rruleset.between(start_datetime, end_datetime, True):
                return True
        except TypeError:
            start_datetime = start_datetime.replace(tzinfo = None)
            end_datetime = end_datetime.replace(tzinfo = None)
            try:
                if rruleset.between(start_datetime, end_datetime, True):
                    return True
            except TypeError:
                return True
        return False
    return True

def match_filter(item, filter):
    if filter is None:
        return True
    if filter.tag != _tag("C", "filter"):
        return True
    for fe in filter:
        if match_filter_element(item.object, fe):
            return True
    return False

def report(path, xml_request, collection):
    """Read and answer REPORT requests.

    Read rfc3253-3.6 for info.

    """
    # Reading request
    root = ET.fromstring(xml_request)

    prop_element = root.find(_tag("D", "prop"))
    props = [prop.tag for prop in prop_element]

    filter_element = root.find(_tag("C", "filter"))

    if collection:
        if root.tag == _tag("C", "calendar-multiget") or root.tag == _tag('A', 'addressbook-multiget'):
            # Read rfc4791-7.9 for info
            hreferences = set((href_element.text for href_element
                               in root.findall(_tag("D", "href"))))
        else:
            hreferences = (path,)
    else:
        hreferences = ()

    # Writing answer
    multistatus = ET.Element(_tag("D", "multistatus"))

    for hreference in hreferences:
        # Check if the reference is an item or a collection
        name = paths.resource_from_path(hreference)
        if name:
            # Reference is an item
            path = paths.collection_from_path(hreference) + "/"
            items = collection.get_items(name)
        else:
            # Reference is a collection
            path = hreference
            items = collection.items


        for item in items:
            if not match_filter(item, filter_element):
                continue

            log.debug("match %s" % item)

            response = ET.Element(_tag("D", "response"))
            multistatus.append(response)

            href = ET.Element(_tag("D", "href"))
            href.text = path.rstrip('/') + '/' + item.name
            response.append(href)

            propstat = ET.Element(_tag("D", "propstat"))
            response.append(propstat)

            prop = ET.Element(_tag("D", "prop"))
            propstat.append(prop)

            for tag in props:
                element = ET.Element(tag)
                if tag == _tag("D", "getetag"):
                    element.text = item.etag
                elif tag == _tag("C", "calendar-data"):
                    element.text = item.text
                elif tag == _tag("A", "address-data"):
                    element.text = item.text
                prop.append(element)

            status = ET.Element(_tag("D", "status"))
            status.text = _response(200)
            propstat.append(status)

    reply = ET.tostring(multistatus, config.get("encoding", "request"))

    return reply



# -*- coding: utf-8 -*-
#
# This file is part of Calypso Server - Calendar Server
# Copyright © 2008-2011 Guillaume Ayoub
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Users and rights management.

This module loads a list of users with access rights, according to the acl
configuration.

"""

from calypso import config


def load():
    """Load list of available ACL managers."""
    acl_type = config.get("acl", "type")
    module = __import__("calypso.acl", fromlist=[acl_type])
    return getattr(module, acl_type)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# This file is part of Calypso Server - Calendar Server
# Copyright © 2009-2011 Guillaume Ayoub
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Calypso.  If not, see <http://www.gnu.org/licenses/>.

"""
Calypso CalDAV server
======================

The Calypso Project is a CalDAV calendar server.  It aims to be a light
solution, easy to use, easy to install, easy to configure.  As a consequence,
it requires few software dependances and is pre-configured to work
out-of-the-box.

The Calypso Project runs on most of the UNIX-like platforms (Linux, BSD,
MacOS X) and Windows.  It is known to work with Evolution 2.30+, Lightning 0.9+
and Sunbird 0.9+. It is free and open-source software, released under GPL
version 3.

For further information, please visit the `Calypso Website
<https://keithp.com/blogs/calypso/>`_.

"""

import os
from distutils.command.build_scripts import build_scripts
from setuptools import setup

try:
    from calypso import VERSION
except ImportError as e:
    print('Error importing Calypso, probably dependencies are not installed')
    print(e)
    VERSION = '0.0.1'
    print('Assuming version %s' % VERSION)

# build_scripts is known to have a lot of public methods
# pylint: disable=R0904
class BuildScripts(build_scripts):
    """Build the package."""
    def run(self):
        """Run building."""
        # These lines remove the .py extension from the calypso executable
        self.mkpath(self.build_dir)
        for script in self.scripts:
            root, _ = os.path.splitext(script)
            self.copy_file(script, os.path.join(self.build_dir, root))
# pylint: enable=R0904


# When the version is updated, ``calypso.VERSION`` must be modified.
# A new section in the ``NEWS`` file must be added too.
setup(
    name="calypso",
    version=VERSION,
    description="CalDAV and CardDAV Server",
    long_description=__doc__,
    author="Keith Packard",
    author_email="keithp@keithp.com",
    url="https://keithp.com/blogs/calypso/",
    download_url="https://anonscm.debian.org/cgit/calypso/calypso.git/",
    license="GNU GPL v3",
    platforms="Any",
    packages=["calypso", "calypso.acl"],
    provides=["calypso"],
    install_requires=["daemon","vobject"],
    tests_require=['nose>=0.11.1'],
    scripts=["calypso.py"],
    cmdclass={"build_scripts": BuildScripts},
    keywords=["calendar", "CalDAV"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Office/Business :: Groupware"])
